/*
 * Saints Row 2 JP Strings Build Tool
 * Created by nclok1405
 *
 * This software uses ThomasJepp.SaintsRow.
 * For more information about ThomasJepp.SaintsRow, see: https://github.com/tomjepp/ThomasJepp.SaintsRow.
 * The license for ThomasJepp.SaintsRow is included with this program in the file "ThomasJepp.SaintsRow license.txt".
 */
#include <iostream>
#include <fstream>
#include <string>
#include <iomanip>
#include <vector>
#include <map>
#include <cstdint>
#include <cctype>
#include <cerrno>
#include <cstring>
#include <utility>

// Raw Structs (can be read from files directly)
struct StringHeader {
    uint32_t id;
    uint16_t version;
    uint16_t bucketCount;
    uint32_t stringCount;
};

struct StringBucket {
    uint32_t stringCount;
    uint32_t stringOffset;
};

// Our data structures (can't be read or written directly)
class StringData {
public:
    uint32_t crc;
    std::string encodedString;
};

class StringBucketData {
public:
    StringBucket bucket;
    std::vector<uint32_t> offsetArray;
    std::vector<StringData> stringDataArray;
};

// Write binary data to an ostream
// https://baptiste-wicht.com/posts/2011/06/write-and-read-binary-files-in-c.html
template<typename T>
std::ostream& binary_write(std::ostream& stream, const T& value){
    return stream.write(reinterpret_cast<const char*>(&value), sizeof(T));
}

int main(int argc, char* argv[])
{
    static const char* INPUT_CHARACTER_TABLE_FILENAME = "sr2_jp_character_table.tbl";
    static const char* INPUT_FILENAME = "static_JP.txt";
    static const char* OUTPUT_FILENAME = "static_JP.le_strings";

    std::string inputFilename(INPUT_FILENAME);
    std::string outputFilename(OUTPUT_FILENAME);

    // Handle the first command line param
    if (argc >= 2) {
        std::string param1(argv[1]);

        if ((param1 == "--help") || (param1 == "-h") || (param1 == "-?") || (param1 == "/help") || (param1 == "/h") || (param1 == "/?") || (param1 == "--version")) {
            // Help
            std::cout << "Saints Row 2 JP Strings Build Tool v0.1" << std::endl;
            std::cout << "Created by nclok1405" << std::endl;
            std::cout << "https://www.saintsrowmods.com" << std::endl;
            std::cout << std::endl;
            std::cout << "This software uses ThomasJepp.SaintsRow." << std::endl;
            std::cout << "For more information about ThomasJepp.SaintsRow, see: https://github.com/tomjepp/ThomasJepp.SaintsRow." << std::endl;
            std::cout << "The license for ThomasJepp.SaintsRow is included with this program in the file \"ThomasJepp.SaintsRow license.txt\"." << std::endl;
            std::cout << std::endl;
            std::cout << "Usage: " << argv[0] << " [Input] [Output]" << std::endl;
            std::cout << "[Input]  : Input Text Filename (Default:" << INPUT_FILENAME << ")" << std::endl;
            std::cout << "[Output] : Output Filename (Default:" << OUTPUT_FILENAME << ")" << std::endl;
            return 0;
        } else {
            // Input Filename
            inputFilename.assign(param1);
        }
    }
    // Handle the second command line param (Output Filename)
    if (argc >= 3) {
        outputFilename.assign(argv[2]);
    }

    // Load the character table
    std::ifstream characterTableInputStream(INPUT_CHARACTER_TABLE_FILENAME, std::ios::binary);
    if (!characterTableInputStream) {
        std::cerr << "ERROR: Failed to load table file '" << INPUT_CHARACTER_TABLE_FILENAME << "' (" << std::strerror(errno) << ")" << std::endl;
        return 1;
    }
    std::map<std::string, uint16_t> characterLookupMap; // Character lookup table
    std::string lineString;

    while (std::getline(characterTableInputStream, lineString)) {
        // Only do lines start with a 0
        if(lineString.find("0") == 0) {
            std::string keyStr = lineString.substr(0, 4);
            uint16_t key = static_cast<uint16_t>(std::stoi(keyStr, nullptr, 16));
            std::string valueStr = lineString.substr(5);
            if (!valueStr.empty()) {
                characterLookupMap[valueStr] = key;
            }
        }
    }

    std::ifstream inputStringFile(inputFilename, std::ios::binary);
    if(!inputStringFile) {
        std::cerr << "ERROR: Failed to load " << inputFilename << " (" << std::strerror(errno) << ")" << std::endl;
        return 2;
    }

    std::vector<StringData> stringDataArray;

    uint32_t linecount = 0;
    while(std::getline(inputStringFile, lineString)) {
        linecount++;
        std::size_t tabPos = lineString.find('\t');

        if (tabPos != std::string::npos) {
            std::string strCRC;
            std::string strBody;
            strCRC = lineString.substr(0, tabPos);
            strBody = lineString.substr(tabPos+1);

            if (!strCRC.empty()) {
                uint32_t crc = 0;
                try {
                    crc = static_cast<uint32_t>(std::stoul(strCRC, nullptr, 16));
                } catch (...) {
                    std::cerr << "WARNING: Cannot convert CRC " << strCRC << " at line " << linecount << std::endl;
                }

                std::size_t p = 0;

                std::string strConverted;

                char crc1 = static_cast<char>((crc & 0xFF000000) >> 24);
                char crc2 = static_cast<char>((crc & 0x00FF0000) >> 16);
                char crc3 = static_cast<char>((crc & 0x0000FF00) >> 8);
                char crc4 = static_cast<char>((crc & 0x000000FF) >> 0);
                strConverted.append(1, crc1);
                strConverted.append(1, crc2);
                strConverted.append(1, crc3);
                strConverted.append(1, crc4);

                p = 0;
                while (p < strBody.size()) {
                    char c = strBody.at(p);
                    uint8_t bytesOfThisCharacter;
                    bool escapeSeq = false;

                    if (c == '\\') {
                        escapeSeq = true;
                        char c2 = strBody.at(p+1);

                        if(c2 == 'n') {
                            // \n
                            strConverted.append(1, '\0');
                            strConverted.append("\n");
                            bytesOfThisCharacter = 2;
                        } else if(c2 == '\t') {
                            // \t
                            strConverted.append(1, '\0');
                            strConverted.append("\t");
                            bytesOfThisCharacter = 2;
                        } else if(c2 == '\\') {
                            // Backslash
                            strConverted.append(1, '\0');
                            strConverted.append("\\");
                            bytesOfThisCharacter = 2;
                        } else if(c2 == '\r') {
                            // \r
                            strConverted.append(1, '\0');
                            strConverted.append("\r");
                            bytesOfThisCharacter = 2;
                        } else if(c2 == 'u') {
                            // \u0123
                            std::string strCodePoint = strBody.substr(p+2, 4);
                            uint16_t codePointValue = 0;
                            try {
                                codePointValue = static_cast<uint16_t>(std::stoul(strCodePoint, nullptr, 16));
                            } catch (...) {
                                std::cerr << "WARNING: Cannot get code point " << strCodePoint << " at line " << linecount << std::endl;
                            }
                            char codePointValue1 = static_cast<char>((codePointValue & 0xFF00) >> 8);
                            char codePointValue2 = static_cast<char>(codePointValue & 0x00FF);
                            strConverted.append(1, codePointValue1);
                            strConverted.append(1, codePointValue2);
                            bytesOfThisCharacter = 6;
                        } else {
                            // ???
                            strConverted.append(1, '\0');
                            strConverted.append("\\");
                            bytesOfThisCharacter = 1;
                        }
                    } else if ((c & 0x80) == 0) {
                        bytesOfThisCharacter = 1;
                    } else if ((c & 0xE0) == 0xC0) {
                        bytesOfThisCharacter = 2;
                    } else if ((c & 0xF0) == 0xE0) {
                        bytesOfThisCharacter = 3;
                    } else {
                        bytesOfThisCharacter = 4;
                    }

                    if (!escapeSeq) {
                        if (bytesOfThisCharacter == 3) {
                            // Japanese
                            std::string strUTF8 = strBody.substr(p, bytesOfThisCharacter);
                            if (characterLookupMap.count(strUTF8)) {
                                uint16_t sr2char = characterLookupMap[strUTF8];
                                char sr2char1 = static_cast<char>((sr2char & 0xFF00) >> 8);
                                char sr2char2 = static_cast<char>(sr2char & 0x00FF);
                                strConverted.append(1, sr2char1);
                                strConverted.append(1, sr2char2);
                            } else {
                                strConverted.append(1, '\0');
                                strConverted.append(1, '?');
                                std::cerr << "WARNING: Unknown character '" << strUTF8 << "' at line " << linecount << std::endl;
                            }
                        } else if (bytesOfThisCharacter == 1) {
                            // 1 byte
                            strConverted.append(1, '\0');
                            strConverted.append(strBody.substr(p, 1));
                        } else {
                            // Others
                            strConverted.append(strBody.substr(p, bytesOfThisCharacter));
                        }
                    }

                    p += bytesOfThisCharacter;
                }

                // End Of String
                strConverted.append(1, '\0');
                strConverted.append(1, '\0');

                StringData stringData;
                stringData.crc = crc;
                stringData.encodedString = strConverted;
                stringDataArray.push_back(stringData);
            }
        }
    }

    // Set header data
    StringHeader header;
    header.id = 0xA84C7F73;
    header.version = 1;
    header.stringCount = static_cast<uint32_t>(stringDataArray.size());

    // Bucket count is 512 in voice_script_JP.le_strings
    // 32 in platform_PC_JP.le_strings
    // stringCount / 5 doesn't produce right bucket count
    header.bucketCount = static_cast<uint16_t>(header.stringCount / 10);
    if (header.bucketCount < 32) {
        header.bucketCount = 32;
    } else if (header.bucketCount < 64) {
        header.bucketCount = 64;
    } else if (header.bucketCount < 128) {
        header.bucketCount = 128;
    } else if (header.bucketCount < 256) {
        header.bucketCount = 256;
    } else if (header.bucketCount < 512) {
        header.bucketCount = 512;
    } else {
        header.bucketCount = 1024;
    }

    // Create string "buckets"
    std::vector<StringBucketData> bucketDataArray(header.bucketCount);

    for (std::size_t i = 0; i < stringDataArray.size(); i++) {
        uint32_t bucketID = static_cast<uint32_t>(stringDataArray.at(i).crc & (header.bucketCount-1));
        bucketDataArray.at(bucketID).stringDataArray.push_back(stringDataArray.at(i));
        bucketDataArray.at(bucketID).offsetArray.push_back(0xFFFFFFFF); // We don't know the offset yet
        bucketDataArray.at(bucketID).bucket.stringCount = static_cast<uint32_t>(bucketDataArray.at(bucketID).stringDataArray.size());
    }

    // Create an output file
    std::ofstream outputFileStream(outputFilename, std::ios::binary);
    if (!outputFileStream) {
        std::cerr << "ERROR: Failed to create '" << outputFilename << "' for writing (" << std::strerror(errno) << ")" << std::endl;
        return 5;
    }

    // Write Header
    binary_write(outputFileStream, header);

    // Write Buckets (offsets need to be fixed later)
    for (std::size_t bucketID = 0; bucketID < bucketDataArray.size(); bucketID++) {
        binary_write(outputFileStream, bucketDataArray.at(bucketID).bucket);
    }
    // Write String Offsets (need to be fixed later)
    for (std::size_t bucketID = 0; bucketID < bucketDataArray.size(); bucketID++) {
        bucketDataArray.at(bucketID).bucket.stringOffset = static_cast<uint32_t>(outputFileStream.tellp());

        for (std::size_t offsetID = 0; offsetID < bucketDataArray.at(bucketID).offsetArray.size(); offsetID++) {
            uint32_t offset = bucketDataArray.at(bucketID).offsetArray.at(offsetID);
            binary_write(outputFileStream, offset);
        }
    }
    // Write Strings
    for (std::size_t bucketID = 0; bucketID < bucketDataArray.size(); bucketID++) {
        for (std::size_t stringID = 0; stringID < bucketDataArray.at(bucketID).stringDataArray.size(); stringID++) {
            bucketDataArray.at(bucketID).offsetArray.at(stringID) = static_cast<uint32_t>(outputFileStream.tellp());

            const std::string& s = bucketDataArray.at(bucketID).stringDataArray.at(stringID).encodedString;
            for (std::size_t i = 0; i < s.size(); i++) {
                char c = s.at(i);
                binary_write(outputFileStream, c);
            }
        }
    }

    // Go back to the beginning after the header
    outputFileStream.seekp(sizeof(StringHeader), std::ios_base::beg);

    // Write Buckets again
    for (std::size_t bucketID = 0; bucketID < bucketDataArray.size(); bucketID++) {
        binary_write(outputFileStream, bucketDataArray.at(bucketID).bucket);
    }
    // Write String Offsets again
    for (std::size_t bucketID = 0; bucketID < bucketDataArray.size(); bucketID++) {
        for (std::size_t offsetID = 0; offsetID < bucketDataArray.at(bucketID).offsetArray.size(); offsetID++) {
            uint32_t offset = bucketDataArray.at(bucketID).offsetArray.at(offsetID);
            binary_write(outputFileStream, offset);
        }
    }

    return 0;
}
