/*
 * Saints Row 2 JP Strings Extraction Tool
 * Created by nclok1405
 *
 * This software uses ThomasJepp.SaintsRow.
 * For more information about ThomasJepp.SaintsRow, see: https://github.com/tomjepp/ThomasJepp.SaintsRow.
 * The license for ThomasJepp.SaintsRow is included with this program in the file "ThomasJepp.SaintsRow license.txt".
 */
#include <iostream>
#include <fstream>
#include <string>
#include <iomanip>
#include <vector>
#include <map>
#include <cstdint>
#include <cctype>
#include <cerrno>
#include <cstring>
#include <set>

// Raw Structs (can be read from files directly)
struct StringHeader {
    uint32_t id;
    uint16_t version;
    uint16_t bucketCount;
    uint32_t stringCount;
};

struct StringBucket {
    uint32_t stringCount;
    uint32_t stringOffset;
};

// Our data structures (can't be read or written directly)
class StringData {
public:
    uint32_t crc;
    std::vector<uint16_t> characterArray;
};

class StringBucketData {
public:
    StringBucket bucket;
    std::vector<uint32_t> offsetArray;
    std::vector<StringData> stringDataArray;
};

// Read binary data from an istream
// https://baptiste-wicht.com/posts/2011/06/write-and-read-binary-files-in-c.html
template<typename T>
std::istream& binary_read(std::istream& stream, T& value){
    return stream.read(reinterpret_cast<char*>(&value), sizeof(T));
}

int main(int argc, char* argv[])
{
    static const char* INPUT_CHARACTER_TABLE_FILENAME = "sr2_jp_character_table.tbl";
    static const char* INPUT_STRINGS_FILENAME = "static_JP.le_strings";
    static const char* OUTPUT_FILENAME = "static_JP.txt";
    static const uint32_t EXPECTED_ID = 0xA84C7F73;
    static const uint16_t EXPECTED_VERSION = 1;

    std::string inputStringsFilename(INPUT_STRINGS_FILENAME);   // Input le_strings Filename
    std::string outputFilename(OUTPUT_FILENAME);                // Output Filename

    // Handle the first command line param
    if (argc >= 2) {
        std::string param1(argv[1]);

        if ((param1 == "--help") || (param1 == "-h") || (param1 == "-?") || (param1 == "/help") || (param1 == "/h") || (param1 == "/?") || (param1 == "--version")) {
            // Help
            std::cout << "Saints Row 2 JP Strings Extraction Tool v0.1" << std::endl;
            std::cout << "Created by nclok1405" << std::endl;
            std::cout << "https://www.saintsrowmods.com" << std::endl;
            std::cout << std::endl;
            std::cout << "This software uses ThomasJepp.SaintsRow." << std::endl;
            std::cout << "For more information about ThomasJepp.SaintsRow, see: https://github.com/tomjepp/ThomasJepp.SaintsRow." << std::endl;
            std::cout << "The license for ThomasJepp.SaintsRow is included with this program in the file \"ThomasJepp.SaintsRow license.txt\"." << std::endl;
            std::cout << std::endl;
            std::cout << "Usage: " << argv[0] << " [Input] [Output]" << std::endl;
            std::cout << "[Input]  : Input le_strings Filename (Default:" << INPUT_STRINGS_FILENAME << ")" << std::endl;
            std::cout << "[Output] : Output Filename (Default:" << OUTPUT_FILENAME << ")" << std::endl;
            return 0;
        } else {
            // Input Filename
            inputStringsFilename.assign(param1);
        }
    }
    // Handle the second command line param (Output Filename)
    if (argc >= 3) {
        outputFilename.assign(argv[2]);
    }

    // Load the character table
    std::ifstream characterTableInputStream(INPUT_CHARACTER_TABLE_FILENAME, std::ios::binary);
    if (!characterTableInputStream) {
        std::cerr << "ERROR: Failed to load table file '" << INPUT_CHARACTER_TABLE_FILENAME << "' (" << std::strerror(errno) << ")" << std::endl;
        return 1;
    }

    std::map<uint16_t, std::string> characterMap; // Character conversion table
    std::set<std::string> characterSet; // Set of characters already used
    std::set<uint16_t> characterKeySet;  // Set of keys already appeared

    uint32_t characterTableLineCount = 0;
    std::string lineString;
    while (std::getline(characterTableInputStream, lineString)) {
        characterTableLineCount++;

        // Only do lines start with a 0
        if(lineString.find("0") == 0) {
            std::string keyStr = lineString.substr(0, 4);
            uint16_t key = static_cast<uint16_t>(std::stoi(keyStr, nullptr, 16));
            std::string valueStr = lineString.substr(5);
            if (!valueStr.empty()) {
                characterMap[key] = valueStr;

                if(characterKeySet.count(key) == 0) {
                    characterKeySet.insert(key);
                } else {
                    std::cerr << "WARNING: Character key '" << keyStr << "' appeared more than once in the table file (line:" << characterTableLineCount << ")" << std::endl;
                }
                if(characterSet.count(valueStr) == 0) {
                    characterSet.insert(valueStr);
                } else {
                    std::cerr << "WARNING: Character '" << valueStr << "' appeared more than once in the table file (line:" << characterTableLineCount << ")" << std::endl;
                }
            }
        }
    }
    // Character Table is done
    characterTableInputStream.close();

    // Load the strings file
    std::ifstream ifs(inputStringsFilename, std::ios::binary);
    if (!ifs) {
        std::cerr << "ERROR: Failed to load '" << inputStringsFilename << "' (" << std::strerror(errno) << ")" << std::endl;
        return 2;
    }

    // Read Header
    StringHeader header;
    binary_read(ifs, header);

    // Sanity Check
    if (header.id != EXPECTED_ID) {
        std::cerr << std::hex << "ERROR: Unexpected Header ID; expected 0x" << EXPECTED_ID << ", got 0x" << header.id << std::dec << std::endl;
        return 3;
    }
    if (header.version != EXPECTED_VERSION) {
        std::cerr << "ERROR: Unexpected Header Version; expected " << EXPECTED_VERSION << ", got " << header.version << std::endl;
        return 4;
    }

    // Read "String Buckets"
    std::vector<StringBucketData> bucketDataVector(header.bucketCount);
    uint32_t stringTotalCount = 0;

    for (uint16_t i = 0; i < header.bucketCount; i++) {
        StringBucket bucket;
        binary_read(ifs, bucket);

        StringBucketData bucketData;
        bucketData.bucket = bucket;
        bucketDataVector.at(i) = bucketData;

        stringTotalCount += bucket.stringCount;
    }

    // Warn if the string counts mismatch
    if (stringTotalCount != header.stringCount) {
        std::cerr << "WARNING: The total string count from buckets and the header's string count are different" << std::endl;
        std::cerr << "String count from all buckets: " << stringTotalCount << std::endl;
        std::cerr << "String count from the header:  " << header.stringCount << std::endl;
    }

    // Read String Data Offsets for each bucket
    for (uint16_t i = 0; i < header.bucketCount; i++) {
        //std::cout << "Bucket #" << i << std::endl;

        StringBucketData& bucketData = bucketDataVector.at(i);
        const uint32_t stringOffset = bucketData.bucket.stringOffset;
        const uint32_t stringCount = bucketData.bucket.stringCount;

        ifs.seekg(stringOffset, std::ios_base::beg);

        for (uint32_t j = 0; j < stringCount; j++) {
            uint32_t offsetToStringData;
            binary_read(ifs, offsetToStringData);
            bucketData.offsetArray.push_back(offsetToStringData);
        }
    }

    // Read the actual strings
    for (uint16_t i = 0; i < header.bucketCount; i++) {
        StringBucketData& bucketData = bucketDataVector.at(i);
        const uint32_t stringCount = bucketData.bucket.stringCount;

        for (uint32_t j = 0; j < stringCount; j++) {
            const uint32_t offsetToStringData = bucketData.offsetArray.at(j);

            ifs.seekg(offsetToStringData, std::ios_base::beg);

            StringData stringData;

            // Read CRC
            uint8_t crc1;
            uint8_t crc2;
            uint8_t crc3;
            uint8_t crc4;
            binary_read(ifs, crc1);
            binary_read(ifs, crc2);
            binary_read(ifs, crc3);
            binary_read(ifs, crc4);
            stringData.crc = static_cast<uint32_t>((crc1 << 24) | (crc2 << 16) | (crc3 << 8) | crc4);

            // Read each character until 00 00 appears
            while (true) {
                uint8_t c1;
                uint8_t c2;
                binary_read(ifs, c1);
                binary_read(ifs, c2);

                if((c1 == 0) && (c2 == 0)) {
                    break;
                }

                uint16_t character = static_cast<uint16_t>((c1 << 8) | c2);
                stringData.characterArray.push_back(character);
            }

            bucketData.stringDataArray.push_back(stringData);
        }
    }

    // Create an output file
    std::ofstream outputFileStream(outputFilename, std::ios::binary);
    if (!outputFileStream) {
        std::cerr << "ERROR: Failed to create '" << outputFilename << "' for writing (" << std::strerror(errno) << ")" << std::endl;
        return 5;
    }

    // Write strings to a file
    for (uint16_t i = 0; i < header.bucketCount; i++) {
        StringBucketData& bucketData = bucketDataVector.at(i);
        const uint32_t stringCount = bucketData.bucket.stringCount;

        for (uint32_t j = 0; j < stringCount; j++) {
            const StringData& stringData = bucketData.stringDataArray.at(j);

            outputFileStream << std::hex << std::uppercase << std::setfill('0') << std::setw(8) << std::right << stringData.crc << std::nouppercase << std::dec << "\t";

            for (std::size_t k = 0; k < stringData.characterArray.size(); k++) {
                const uint16_t character = stringData.characterArray.at(k);

                if(characterMap.count(character) != 0) {
                    // Put a character from the table
                    outputFileStream << characterMap[character];
                } else if(character == '\\') {
                    // Backslash
                    outputFileStream << "\\\\";
                } else if(character == '\n') {
                    // Newline
                    outputFileStream << "\\n";
                } else if(character == '\t') {
                    // Tab
                    outputFileStream << "\\t";
                } else if(character == '\r') {
                    // \r
                    outputFileStream << "\\r";
                } else if( (character <= 0x007F) && std::isprint(character) ) {
                    // Printable ASCII Characters
                    const char c = static_cast<char>(character);
                    outputFileStream << c;
                } else {
                    // Unknown/Missing Characters
                    outputFileStream << "\\u" << std::hex << std::uppercase << std::setfill('0') << std::setw(4) << std::right << character << std::nouppercase << std::dec;
                }
            }

            outputFileStream << std::endl;
        }
    }

    return 0;
}
